// SpMV sparse-matrix-vector-multiplication benchmark (driven by rnd-mem-access)
//  (2nd version) hybrid code test vs. OMP-test (derived from memtest.c)
//  - single core version works (speed test2 5..15% slower than test1)
//  - OpenMP works for test=1 only (test2 has race conditions, missing sort)
//  - MPI does not work at the moment (ToDo 2017-02)
const char *version_string="2017-02-26";
// ToDo: use _OPENMP macro autoset by -fopenmp (grep -e omp -e openmp *.c)
// // #define CFG_OMP  /* obsolete! use: gcc -fopenmp ... */
// #define CFG_MPI        /* define CFG_MPI + mpicc ... */ 
// #define CFG_CPUSET     /* options --pin_ncpu=... */
#if defined(_OPENMP) && !defined(CFG_OMP)  /* 2016-01-14 */
# define CFG_OMP
#endif
/*
 also FDIR
  - Ausfaelle erkennen (fault-detection)          
  - Ausfaelle einkreisen und bestimmen (fault-isolation)
  - geeignete Korrekturen (recovery)    
 
 INST_SEQ: # single task (no parallelism)
          gcc -Wall -O2 -lm -o memspeed memspeed.c
          # memory footprint: 
 INST_MPI: # OpenMPI-usage (parallel on distributed memory)
          OMPI_CC=gcc492 mpicc -DCFG_MPI -Wall -O2 -lm \
           -ftree-vectorize -funroll-loops \
           -o memspeed memspeed.c
          # optional: -DCFG_CPUSET=0x0ff -fopenmp
          #   cpuset is a hex bit pattern for used cpu cores (p.e. 0x05,0x0f)
          #     useful to bind tasks to cores using sched_setaffinity()
          #   use cc or gcc if mpicc is not availabe
          mpirun -bind-to-core -np <tasks>\
            ./memspeed --asize=<MBsize> --mintime=<sec>
          # needs: memory = tasks*2^log2size Bytes
          #        time   = (1..2)*2*mintime*(log2size-8) seconds
          mpirun -np 8 --bynode ./memspeed --log2size=29 #    4GB=8*512MB
          mpirun -np 2 --bynode ./memspeed --log2size=27 #  256MB=2*128MB
          mpirun -np 2 --bynode ./memspeed --asize=384   #  768MB=2*384MB
          # runtime is about 2*(minsize...2minsize)*(log2size-7)
          ... --pin_ncpu=6 ...
 INST_OMP:# OpenMP-usage (parallel on shared memory)
          yum install libgomp
          gcc -Wall -O2 -fopenmp -lm -o memspeed memspeed.c
          export OMP_WAIT_POLICY="PASSIVE" # OpenMP-3.0 waits saving power 
          export OMP_NUM_THREADS=8         # OpenMP     8 threads
          OMP_PROC_BIND=TRUE # thread bounding, thread affinity
          OMP_WAIT_POLICY="PASSIVE" OMP_NUM_THREADS=2 ./memspeed
          # memory footprint: 
 SLURM:   # job-system
          # task/affinity or task/cgroup must be enabled, scontrol show config
          # SLURM_CPU_BIND=verbose # -c = --cpus-per-task (OpenMP) dflt=1
          srun -N1 -c4 --cpu_bind=verbose,mask_cpu:0xf,0xf0,0xf00,0xf000 ./memspeed --asize=384
          srun -N1 --ntasks-per-node=4 --cpu_bind=verbose,mask_cpu:0xf,0xf0,0xf00,0xf000
          # -u = --unbuffered (output)
          #  SLURM_CPU_BIND_VERBOSE SLURM_CPU_BIND_TYPE SLURM_CPU_BIND_LIST is set
          # srun works with mpich, for openmpi mpirun musst be called:
          srun -N1-2 -c8 bash -c "[ \$SLURM_PROCID == 0 ]&&mpirun ./memspeed||true"
          srun -p sw01_short -N1-2 -c8 bash -c "[ \$SLURM_PROCID = 0 ]&&OMP_NUM_THREADS=4 mpirun ./memspeed --test=1 || true"

ToDo:
 - use much faster random_xorshift()
 - random_update vy[prefetchable_rnd_iy]+=nzHxy[rnd_iy]*vx[seq_ix]
   from rnd_nz_matrix p.e. NZX=40; vx,vy,sym_sparse_matrix[iy,ry] SpMV
  CSR = compressed sparse row, 
   large.N>=1e5 nnz/row=10..100 100-300MFLOP/s
  Benchmarking_Sparse_Matrix_Vector.pdf
    diag1: x=N y=nnz/row color=MFLOP/s
    diag2: x=N y=MFLOP/s color=nnz/row(fix to log(N)?)

   old16: [iy,Hxy] snd[iy]       -    rcv[vy] vx+=Hxy*vy(seq_write,rnd_rdbuf)
    remote:   -    rcv[iy] rnd_rd[vy] snd[vy] -
   new18? [iy,Hxy] Hxy*vx snd[iy,Hxy*vx]      -        # more simple, faster?
    remote:   -           rcv[iy,Hxy*vx] rnd_vy+=Hxy*vx
    
 - use vector-code (SSE2, AVX2) for speedup
ChangeLog:
 2017-02-27: initial version from memtest.c (to check access-pattern-versions)
*/


#define _GNU_SOURCE 1 // this defines __USE_GNU in features.h for sched.h

#ifdef _OPENMP
#include <omp.h>    /* yum install libgomp; gcc -fopenmp ... */
#endif
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* 2013-05 adding strncmp() */
#include <time.h>
#include <unistd.h> /* usleep */
#include <math.h> /* sqrt */

#ifndef CFG_MPI    // for the case of MPI is not available
/* last tests: 2013-05-02 */
#include <sys/time.h>
// ToDo: add Ialltoallv
double wtime(){ struct timeval tv;
 if(gettimeofday(&tv, NULL)){perror("timeofday");exit(2);}
 return tv.tv_sec+1e-6*tv.tv_usec; }
#define MPI_Init(a,b) 1
#define MPI_SUCCESS 1
#define MPI_Comm_size(MPI_COMM_WORLD, mpi_n)
#define MPI_Comm_rank(MPI_COMM_WORLD, mpi_i)
#define MPI_Barrier(MPI_COMM_WORLD)
#define MPI_Wtime()  wtime() /* 2013-05-02 itime replaced by wtime */
#define MPI_Finalize()
#define MPI_Abort(MPI_COMM_WORLD,xx) exit(xx)
#define MPI_Bcast(pt, len, MPI_DOUBLE, id, MPI_COMM_WORLD)
/* Reduce replacement works for MIN MAX SUM,
   Reduce to rank0, Allreduce to all ranks */
#define MPI_Reduce(snd, rec, SIZE, M_INT, M_MIN, rk, MCW) *rec=*snd
#define MPI_Allreduce(snd, rec, SIZE, M_INT, M_MIN, MCW) *rec=*snd
#else  // CFG_MPI
#include <mpi.h>
#endif // CFG_MPI

#ifdef CFG_CPUSET // switch on if sched_setaffinity will benefit
#include <unistd.h>  // declare sleep()+usleep
#include <sched.h>   // declare sched_setaffinity()
#ifndef _FEATURES_H  
# warning "_FEATURES_H is not defined, try top add #include <features.h>"
#endif // _FEATURES_H
#ifndef __USE_GNU
// bcc 0.16.17 does not like #warning?
# warning "expect __USE_GNU defined, but it isn't, try -D_GNU_SOURCE=1"
#endif // __USE_GNU
#endif // CFG_CPUSET

#define ULong  unsigned long int

/* very fast pseudo random generator
 *   (replacement of slow urandom, fake-flash-test, JoergSchbg 2016)
 *   see https://en.wikipedia.org/wiki/Xorshift
 *  hpmini-gcc47-1GHz: urandom=1.8MB/s xorshift1024*=133MB/s
 *  JPC2010-gcc492:    urandom=4.0MB/s xorshift1024*=920MB/s gcc41=940MB/s
 */
/* The state must be seeded so that it is not everywhere zero. */
uint64_t s[16]={5,28,37,23,1234,2,12334,3987,873,321,456,533245,24,455,22,23};
int p=0; // not thread save!

uint64_t xorshift(){
  uint64_t rnd;
  /* rnd= xorshift1024star(); */
  const uint64_t s0 = s[p];
  uint64_t s1 = s[p = (p + 1) & 15];
  s1 ^= s1 << 31; // a
  s[p] = s1 ^ s0 ^ (s1 >> 11) ^ (s0 >> 30); // b, c
  rnd = s[p] * UINT64_C(1181783497276652981);   
  return rnd;
}

#define NZX  40              /* 8..100 */
double *vx, *vy, *hxy;
ULong *hiy; /* h = nzH */
ULong *ofsy;  // ofsets to hxy-Matrix (trivial for const. NZX-steps?)

int main(int argn, char *argv[]){
  double t1, t2, tmin=1.0, rr=0.0;
  int omp_n=1,      /* OpenMP omp_num_threads() */
      mpi_n=1, mpi_i=0, /* MPI threads and thread number */
      pin_flag=0, /* thread to processor pinning */
      testmode=  1, /* 1=seq_read 2=pf_random 4=nonpf_random 8=stress ... */
      log2size= 20; // 2^20=1MB  set by argument1 (old, new: unit)
  long int loops, asize= 2 /* 2<<20 * (1+1+1+40)*8B= 4M*344B = 0.69GB */,
     i1, i4, i, j; // dummy vars for the innerst loop (unique names!)
  if (MPI_Init(&argn, &argv) != MPI_SUCCESS) exit(1);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_n);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_i);
  setvbuf(stdout, NULL, _IONBF, 0); /* dont buffer output */
  if (!mpi_i)
   printf("# SpMV-speed %s (c) 2017-2017 GPLv2 Joerg Schulenburg ovgu.de\n",
           version_string);
  /* 2013-05 add option strings --... for more flexibility */
  for (i1=1;i1<argn;i1++) { /* check all args */
    if (strncmp(argv[i1],"--asize=",8)==0) { asize=atoi(argv[i1]+8);
      if (asize<1) asize=1; continue; } // min 1 * 2^log2size 
    if (strncmp(argv[i1],"--log2size=",11)==0) { log2size=atoi(argv[i1]+11);
      continue; } // min 1K?
    if (strncmp(argv[i1],"--mintime=",10)==0) { 
      tmin=atof(argv[i1]+10); if (tmin<0.0) tmin=0.0; continue; }
    if (strncmp(argv[i1],"--test=",7)==0) { 
      testmode =atoi(argv[i1]+7); continue; }
    /* pin_* = processor core/thread pinning */
    if (strncmp(argv[i1],"--pin_ncpu=",11)==0) { pin_flag=1; continue; }
    if (strncmp(argv[i1],"--pin_ofs=", 10)==0) { pin_flag=1; continue; }
    if (strncmp(argv[i1],"--pin_dist=",11)==0) { pin_flag=1; continue; }
    if (strncmp(argv[i1],"--pin_size=",11)==0) { pin_flag=1; continue; }
    if (!mpi_i) {
     printf("ERROR: bad or unknown options or arguments\n"
        " use: ./SpMV_speed --asize=16 --log2size=20 --mintime=1 --test=1\n");
     fputs(" --pin_ncpu=16 --pin_ofs=0 --pin_dist=1 --pin_size=1\n",stdout);
     fputs(" --test=1  [iy,Hxy] snd[iy] rnd_rd[vy] rcv[vy] vx+=Hxy*vy\n",stdout);
     fputs(" --test=2  [iy,Hxy] Hxy*vx snd[iy,Hxy*vx] rnd_vy+=Hxy*vx\n",stdout); /* pf_ */
    }
    usleep(100000l); /* give 100ms for output */ MPI_Abort(MPI_COMM_WORLD,1);
  }
  asize= asize << log2size;

#if defined(CFG_CPUSET) && defined(__USE_GNU)
  if (pin_flag==1) { /* processor/thread pinning */
   int j1;
   cpu_set_t cpuset; // typedef struct { ULong bits[CPU_SETSIZE / NCPUBITS]; }
   struct s_pin {int ncpu;int ofs;int dist;int size;}; // new2013-05-07
   int  ncores=0, ii; unsigned long cmask=0L;
   struct s_pin pin = {16,0,1,1}; // CFG_CPUSET=16,0,1,1 mask=0xffff
   // struct s_pin pin = {8,1,4,2}; // mask=0x00066
   for (i1=1;i1<argn;i1++) { /* check all args */
      if (strncmp(argv[i1],"--pin_ncpu=",11)==0) {
         pin.ncpu=atoi(argv[i1]+11); continue; }
      if (strncmp(argv[i1],"--pin_ofs=",10)==0) {
         pin.ofs= atoi(argv[i1]+10); continue; }
      if (strncmp(argv[i1],"--pin_dist=",11)==0) {
         pin.dist=atoi(argv[i1]+11); continue; }
      if (strncmp(argv[i1],"--pin_size=",11)==0) {
         pin.size=atoi(argv[i1]+11); continue; }
   }
   if (pin.ncpu<1) {printf("ERROR: pin.ncpu<1\n");exit(1);}
   if (pin.size<1) {printf("ERROR: pin.size<1\n");exit(1);}
   if (pin.dist<0) {printf("ERROR: pin.dist<1\n");exit(1);}
   if (pin.ofs>=pin.ncpu) {printf("ERROR: pin.shift\n");exit(1);}
   // auto-detect num cores by try+fail sched_affinity
   CPU_ZERO(&cpuset); 
   for(j1=0,i=CPU_SETSIZE;i;i>>=1){
     CPU_SET(j1|i,&cpuset);
     if (!sched_setaffinity(0, sizeof(cpu_set_t), &cpuset)) j1|=i; // no_err!
     CPU_CLR(j1|i,&cpuset);
   }
   if (!mpi_i) printf("# Max-detected-CPU-Cores/Threads= %d\n", j1+1);
   ncores=j1+1;
      
   for (ii=0;ii<8*sizeof(long) && ii<mpi_n;ii++) {  // set cmask
     j1=((ii/pin.size)*pin.dist
        +(ii%pin.size)+pin.ofs)%pin.ncpu;
     if (j1<8*sizeof(long)) cmask |= 1UL<<j1;
   }
   if (!mpi_i)
     printf("# CPU-Pinning: idx=((i/%d)*%d+(i%%%d)+%d)%%%d, mask=0x...%08lx\n",
     pin.size, pin.dist, pin.size, pin.ofs, pin.ncpu,
      cmask);
   if (pin.ncpu>ncores && !mpi_i) printf("WARNING: pin.ncpu>ncores\n");
   // does only cover equidistant distribution but most
   // MASK     
   // 0xffff = {16,1,1,0}
   // 0x6666 = {16,2,4,1}
   // 0x0f0f = {16,4,8,0} = ((ii/4)*8+0)%16
   // 0xf0f0 = {16,4,8,4}
   ii=((mpi_i/pin.size)*pin.dist
      +(mpi_i%pin.size)+pin.ofs)%pin.ncpu;
   // ncores=pin.ncpu;
   if (ii>=CPU_SETSIZE) {printf("ERROR: ii>CPU_SETSIZE, ii=%d\n",ii);exit(1);}
   CPU_ZERO(&cpuset);   // void CPU_ZERO(cpu_set_t *set); bits[idx]|=mask
   CPU_SET(ii,&cpuset); //  ->bits[ii/NCPUBITS] |= (1<<(ii%NCPUBITS))
   
   // The constant CPU_SETSIZE (1024) specifies a value one greater than  the
   //      maximum CPU number that can be stored in a CPU set.
   usleep((mpi_i%16+1)*10000L); // 10ms..170ms, sort outputs
   if (mpi_i<4 || mpi_i>=mpi_n-2) {
     printf("# set cpu affinity for core(%d)=%d size=%d\n",
              mpi_i, ii, CPU_SETSIZE);
   }
   // no sched_setaffinity on OSF1
   if (sched_setaffinity(0, sizeof(cpu_set_t), &cpuset)) {
     printf("# set cpu affinity for core(%d)=%d size=%d failed\n",
             mpi_i, ii, CPU_SETSIZE);
             usleep(100000L);perror("set_affinity");
   }
   usleep(100000); // 100ms, pseudo sync
  }
  MPI_Barrier(MPI_COMM_WORLD);
#endif
  if (!mpi_i) printf("#");
   /* 2013-05 gcc -fopenmp -v ... =
     -L/usr/lib/gcc/x86_64-redhat-linux/4.1.2 -lgomp -lpthread */
#ifdef _OPENMP
  // omp_set_num_threads(2);
#pragma omp parallel
  omp_n= omp_get_num_threads();  /* must be called within parallel pragma */
  /* next line is outside parallel region */
  if (!mpi_i) printf(" OpenMP threads=%d", omp_n); /* same on every node? */
#endif
#ifdef MPI_VERSION
  if (!mpi_i) printf(" MPI%d.%d",MPI_VERSION,MPI_SUBVERSION);
#endif
  if (!mpi_i)
    printf(" mpitasks=%d size/rank=%.1fM*flt8B mintime/s=%.1f test%d\n"
           "# measure: BW = bandwidth, aBW = aggregate BW, lat = latency\n",
      mpi_n, asize/1000000., tmin, testmode);

//double *vx, *vy, (*hxy)[NZX];
//ULong *ofsy;  // ofsets to hxy-Matrix (trivial for const. NZX-steps?)
//ULong vlen=0;  // vector length (matrix dimension)
  
  vx =  (double *) malloc( asize   *sizeof(double));
  vy =  (double *) malloc( asize   *sizeof(double));
  ofsy = (ULong *) malloc((asize+1)*sizeof(ULong)); /* asize+1 for LEN */
  hxy=  (double *) malloc((asize+asize/10+10)*sizeof(double)*(NZX)); /* +reserve */
  hiy=   (ULong *) malloc((asize+asize/10+10)*sizeof(ULong)*(NZX));
  { /* MPI_Allreduce !aa and get min/max maxmalloc, also sync 2013-06 */
    int aa1, aa2;
    aa1=((vx && vy && ofsy && hxy && hiy)?0:1); /* 1 = failure */
    MPI_Allreduce(&aa1, &aa2, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
    if (aa2 && !mpi_i) printf("ERROR: malloc failed on %d nodes\n", aa2);
    if (aa1 && aa2<8) printf("INFO: malloc failed on node %d\n", mpi_i);
    if (aa2) { exit(1); }
  }
  { /* preinit a_rnd and reuse for speedup, but randomness gets more bad */
    long int i1,i2,i3,ix,iy; ULong rnd;
    // srandom(mpi_i);  // different access patterns on different CPUs
    // ToDo: OMP for thread locality!
    printf("# initialize spare matrix and vectors dim=%lu nzx=%d\n",
             asize, NZX);
#pragma omp parallel for default(shared) private(i1) firstprivate(asize)
    for (i1=0;i1<asize;i1++) vx[i1]=0.0; vx[0]=0.0;
#pragma omp parallel for default(shared) private(i1) firstprivate(asize)
    for (i1=0;i1<asize;i1++) vy[i1]=0.0; vy[0]=1.0;
#pragma omp parallel for default(shared) private(i1) firstprivate(asize)
    for (i1=0;i1<asize+1;i1++) ofsy[i1]=i1*NZX;
#pragma omp parallel for default(shared) private(i1,i2) firstprivate(asize)
    for (i1=0;i1<asize;i1++) {
      for (i2=0;i2<NZX;i2++) {
        hxy[i1*NZX+i2]=0.0; 
        hiy[i1*NZX+i2]=i1;
      } }
    if (asize<1) exit(1);
    if (asize<=NZX) exit(2); // endless loop
    hxy[    0]=-2.0; hiy[    0]=0; /* dominant diagonal element */
    hxy[NZX  ]=-1.0; hiy[NZX  ]=1; /* dominant diagonal element */
    hxy[    1]= 1.0; hiy[    1]=1; /* dominant non-diagonal element */
    hxy[NZX+1]= 1.0; hiy[NZX+1]=0; /* dominant non-diagonal element */
//#pragma omp parallel for default(shared) private(i1,i2) firstprivate(asize)
    for (i1=0;i1<asize;i1++) {
      for (i4=i2=0;i2<NZX && i4<4*NZX;i4++) {
        if (hxy[i1*NZX+i2]!=0.0){ i2++;continue; }
        rnd=xorshift();
        rr=(rnd&7)/4.-1.0; if (rr>=0) rr+=1/4.; rnd>>=3;
        iy=i1 + rnd%((asize-i1)/2+1);
        for (i3=0;i3<NZX;i3++) // multiple [ix,iy]?
          if (hiy[i1*NZX+i3]==iy && hxy[i1*NZX+i3]!=0.0) break;
        //if (i1<4)
        //  printf("DBG.ini x=%5ld nzx=%3ld %3ld %5ld iy=%5ld %6.3f\n",
        //    i1,i2,i3,i4, iy,rr);
        if (i3<NZX) continue; /* no 2 elements with same row */
        for (i3=0;i3<NZX;i3++)  if (hxy[iy*NZX+i3]==0.0) break; // empty slot
        if (i3>=NZX) continue; /* no empty slot */
        //if (i1<4)
        //  printf("DBG.set x=%5ld nzx=%3ld %3ld %5ld iy=%5ld %6.3f ++\n",
        //    i1,i2,i3,i4, iy,rr);
        hxy[i1*NZX+i2]=rr; 
        hiy[i1*NZX+i2]=iy;
        hxy[iy*NZX+i3]=rr;
        hiy[iy*NZX+i3]=i1;
        i2++;
      } }
    printf("# initialization finished ix=%lu ofsy[asize]=%lu max=%lu\n",
       asize, ofsy[asize], (asize+asize/10+10)*NZX);
    // make sparse matrix symmetric! (not parallel!?) or randomize
    i4=0;
    for (ix=i1=0;i1<asize;i1++) for (i2=0;i2<NZX;i2++) {
       ix = i1;
       iy = hiy[ i1*NZX+i2 ];
       rr = hxy[ i1*NZX+i2 ];
       if (ix<8 && iy<8 && asize<=2*NZX)
         printf("nzH: i,ix,iy,hxy %3lu %5lu %5lu %7.4f\n",
           i1*NZX+i2, ix, iy, rr);
       //if (i1==0 && i2<2) continue;
       if (rr==0.0) continue;
       for (i3=0;i3<NZX;i3++) // search symmetric entry (+mark?)
         if (hiy[iy*NZX+i3]==ix && hxy[iy*NZX+i3]==rr) break;
       if (i3==NZX && i4<16){ i4++;printf(" sym %5lu.%2lu not found\n",i1,i2); }
    }
    printf("# initialization check finished\n");
    if (ofsy[ix] > (asize+asize/10+10)*NZX) printf("ERROR nz overflow\n");
  }

#ifndef NA      /* use option -DNA=5 to modify inner loops */
#define NA (8)  /* vector array */
#endif

    if (testmode<1 || testmode>2) {
       printf("bad testmode %d\n",testmode);exit(4);}
    if (!mpi_i)
      printf("# mpi omp dim/Task    loops t[s] aBW[nz/s] BW[nz/s] lat[ns] m=2\n");
    { /* 2014-06-26 start with 8*128kB + go up */
    long int i1, i2; // OMP does not like unsigned
    double rr=0.0, norm2; 
//#pragma omp parallel for default(shared) private(k,i)
    MPI_Barrier(MPI_COMM_WORLD);
    loops=0;
    t1=MPI_Wtime();
    for (i=1;i<(1l<<30);i*=2) {
      j=i;
      if (testmode==1)
      do { // repeat until some seconds are gone
#pragma omp parallel for default(shared) private(i1,i2) firstprivate(asize)
         for (i1=0;i1<asize;i1++) { // vx += H[xy]*vy
           for (i2=ofsy[i1];i2<ofsy[i1+1];i2++) { // one row
             vx[i1] += hxy[i2] * vy[ hiy[i2] ];  // seq_update random read
             // on MPI 2 transfers! hiy to remote + vy[hiy] back
           } } norm2=0.0; rr=0.0;
#pragma omp parallel for private(i1) firstprivate(asize) reduction(+:rr,norm2)
         for (i1=0;i1<asize;i1++){
           rr+=vy[i1]*vx[i1]; norm2+=vx[i1]*vx[i1];}
         norm2=1./sqrt(norm2);
#pragma omp parallel for private(i1) firstprivate(asize)
         for (i1=0;i1<asize;i1++) { vy[i1]=norm2*vx[i1]; vx[i1]=0.0; }
      } while(--j);
      if (testmode==2)
      do { // repeat until some seconds are gone
#pragma omp parallel for default(shared) private(i1,i2) firstprivate(asize)
         for (i1=0;i1<asize;i1++) { // vx += H[xy]*vy
           for (i2=ofsy[i1];i2<ofsy[i1+1];i2++) { // one row
#pragma omp critical
              { vx[hiy[i2]] += hxy[i2] * vy[ i1 ]; } // random update
             // on MPI 1 transfer! hiy + hxy*vy
           } } norm2=0.0; rr=0.0;
#pragma omp parallel for private(i1) firstprivate(asize) reduction(+:rr,norm2)
         for (i1=0;i1<asize;i1++){
           rr+=vy[i1]*vx[i1]; norm2+=vx[i1]*vx[i1];}
         norm2=1./sqrt(norm2);
#pragma omp parallel for private(i1) firstprivate(asize)
         for (i1=0;i1<asize;i1++) { vy[i1]=norm2*vx[i1]; vx[i1]=0.0; }
      } while(--j);               // gcc-4.1 x86 2 instr.
      loops+=i;  // long int = 64bit
      t2=MPI_Wtime()-t1;
      // if (t2>0.1) printf(" loops,erg= %9lu %.6e\n",loops,rr);
      MPI_Bcast(&t2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
      if ((t2>=tmin || t2>10 || loops>>30) && !mpi_i) // alive and final message
      printf("%c%3d %2d %8.2e %9ld %5.2f %8.2e %8.2e %6.2f t%d.nzx%d %7.4f\n",
         ((t2>=tmin)?' ':'#'), mpi_n, omp_n, asize*1.,
         loops, t2, mpi_n*ofsy[asize]*1.*loops/t2,
                          ofsy[asize]*1.*loops/t2, 
                  t2*1e9/(ofsy[asize]*1.*loops), testmode, NZX, rr);
      // l must be used, some compilers remove the mem access else
      if (t2>=tmin) break;
    }
    /* same end-vector? */
//    for (rr=0.0,i1=0;i1<asize;i1++) 
//      if (rr<vy[i1]*vy[i1]) {
//        rr=vy[i1]*vy[i1]; printf(" vy_max %8lu %.4e\n",i1,rr); }
  } /* for (a2=alen;a2>>5;a2>>=1) down to 2**5=32B */

//#pragma omp parallel for default(shared) private(k,ap) reduction(^:x1,x2,x3,x4) firstprivate(aa,a2)
//#pragma omp parallel default(shared) private(k,k2) firstprivate(a2,ry,rb,rc)

  if (!mpi_i) printf("# SpMV_speed finished\n");
  
  MPI_Finalize();
  return (int)rr;
}
